--CRB_Basekit:kitIProgBar_Breath_Base
--CRB_Basekit:kitAccent_Frame_OrangeCorners
--CRB_Basekit:kitAccent_Frame_OrangeCorners
-----------------------------------------------------------------------------------------------
-- Client Lua Script for Nameplates_Remix
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------

require "Window"
require "ChallengesLib"
require "Unit"
require "GameLib"
require "Apollo"
require "PathMission"
require "Quest"
require "Episode"
require "math"
require "string"
require "DialogSys"
require "PublicEvent"
require "PublicEventObjective"
require "CommunicatorLib"
require "GroupLib"
require "PlayerPathLib"
require "GuildLib"
require "GuildTypeLib"

local Nameplates = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local karDisposition =
{
	tTextColors =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= ApolloColor.new("DispositionHostile"),
		[Unit.CodeEnumDisposition.Neutral] 	= ApolloColor.new("DispositionNeutral"),
		[Unit.CodeEnumDisposition.Friendly] = ApolloColor.new("DispositionFriendly"),
	},

	tTargetPrimary =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "Red",
		[Unit.CodeEnumDisposition.Neutral] 	= "xkcdAmber",
		[Unit.CodeEnumDisposition.Friendly] = "Green",
	},

	tHealthBar =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "xkcdRed",
		[Unit.CodeEnumDisposition.Neutral] 	= "xkcdOrange",
		[Unit.CodeEnumDisposition.Friendly] = "xkcdTrueGreen",
	},

	tHealthTextColor =
	{
		[Unit.CodeEnumDisposition.Hostile] 	= "ffff8585",
		[Unit.CodeEnumDisposition.Neutral] 	= "ffffdb57",
		[Unit.CodeEnumDisposition.Friendly] = "ff9bff80",
	},
}


local karConColors =  -- differential value, color
{
	{-4, ApolloColor.new("ConTrivial")},
	{-3, ApolloColor.new("ConInferior")},
	{-2, ApolloColor.new("ConMinor")},
	{-1, ApolloColor.new("ConEasy")},
	{0, ApolloColor.new("ConAverage")},
	{1, ApolloColor.new("ConModerate")},
	{2, ApolloColor.new("ConTough")},
	{3, ApolloColor.new("ConHard")},
	{4, ApolloColor.new("ConImpossible")}
}

local karPathSprite =
{
	[PlayerPathLib.PlayerPathType_Soldier] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSol",
	[PlayerPathLib.PlayerPathType_Settler] 		= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSet",
	[PlayerPathLib.PlayerPathType_Scientist] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathSci",
	[PlayerPathLib.PlayerPathType_Explorer] 	= "CRB_TargetFrameRewardPanelSprites:sprTargetFrame_PathExp",
}

local knTargetRange 		= 40000 -- the distance^2 that normal nameplates should draw within (max targeting range)

-- Todo: break these out onto options
local kcrUnflaggedGroupmate				= ApolloColor.new("DispositionFriendlyUnflagged")
local kcrUnflaggedGuildmate				= ApolloColor.new("DispositionGuildmateUnflagged")
local kcrUnflaggedAlly					= ApolloColor.new("xkcdBlueyGreen")
local kcrFlaggedAlly					= ApolloColor.new("DispositionFriendly")
local kcrUnflaggedEnemyWhenUnflagged 	= ApolloColor.new("DispositionNeutral")
local kcrFlaggedEnemyWhenUnflagged		= ApolloColor.new("DispositionPvPFlagMismatch")
local kcrUnflaggedEnemyWhenFlagged		= ApolloColor.new("DispositionPvPFlagMismatch")
local kcrFlaggedEnemyWhenFlagged		= ApolloColor.new("DispositionHostile")
local kcrDeadColor 						= ApolloColor.new("crayGray")

local kcrDefaultTaggedColor = ApolloColor.new("crayGray")

-- Control types
-- 0 - custom
-- 1 - single check
local Allsave = {}
local currentspec = "1"
local karSavedProperties =
{
	--General nameplate drawing
	["bShowMainObjectiveOnly"] = { default=true, nControlType=1, strControlName="MainShowObjectives" },
	["bShowMainGroupOnly"] = { default=true, nControlType=1, strControlName="MainShowGroup" },
	["bShowMyNameplate"] = { default=true, nControlType=1, strControlName="MainShowMine" },
	["bShowOrganization"] = { default=false, nControlType=1, strControlName="MainShowOrganization" },
	["bShowVendor"] = { default=true, nControlType=1, strControlName="MainShowVendors" },
	["bShowTaxi"] = { default=true, nControlType=1, strControlName="MainShowTaxis" },
	["bShowDispositionHostile"] = { default=true, nControlType=1, strControlName="MainShowDisposition_1" },
	["bShowDispositionNeutral"] = { default=false, nControlType=1, strControlName="MainShowDisposition_2" },
	["bShowDispositionFriendly"] = { default=false, nControlType=1, strControlName="MainShowDisposition_3" },
	["bShowDispositionFriendlyPlayer"] = { default=true, nControlType=1, strControlName="MainShowDisposition_FriendlyPlayer" },
	["bShowIconsDebuff"] = { default=true, nControlType=1, strControlName="MainShowDisposition_Debuff"},
	["bshowHealIcons"] = { default=true, nControlType=1, strControlName="MainShowDisposition_Heal" },
	--Draw distance
	["nMaxRange"] = { default=70.0, nControlType=0 },
	--Individual
	["bShowNameMain"] = { default=true, nControlType=1, strControlName="IndividualShowName", fnCallback="OnSettingNameChanged" },
	["bShowTitle"] = { default=false, nControlType=1, strControlName="IndividualShowAffiliation", fnCallback="OnSettingTitleChanged" },
	["bShowCastBarMainHostile"] = { default=true, nControlType=1, strControlName="IndividualShowCastBarHostile" },
	["bShowCastBarMainFriendly"] = { default=false, nControlType=1, strControlName="IndividualShowCastBarFriendly" },
	["bShowRewardsMain"] = { default=true, nControlType=1, strControlName="IndividualShowRewardIcons", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowLevel"] = { default=false, nControlType=1, strControlName="IndividualShowLevel", fnCallback="OnSettingNameChanged" },
	
	--Reward icons
	["bShowRewardTypeQuest"] = { default=true, nControlType=1, strControlName="ShowRewardTypeQuest", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRewardTypeMission"] = { default=true, nControlType=1, strControlName="ShowRewardTypeMission", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRewardTypeAchievement"] = { default=false, nControlType=1, strControlName="ShowRewardTypeAchievement", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRewardTypeChallenge"] = { default=true, nControlType=1, strControlName="ShowRewardTypeChallenge", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRewardTypeReputation"] = { default=false, nControlType=1, strControlName="ShowRewardTypeReputation", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRewardTypePublicEvent"] = { default=true, nControlType=1, strControlName="ShowRewardTypePublicEvent", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowRivals"] = { default=true, nControlType=1, strControlName="ShowRewardTypeRival", fnCallback="RequestUpdateAllNameplateRewards" },
	["bShowFriends"] = { default=true, nControlType=1, strControlName="ShowRewardTypeFriend", fnCallback="RequestUpdateAllNameplateRewards" },
	--Info panel
	["bShowHealthMain"] = { default=false, nControlType=0, fnCallback="OnSettingHealthChanged" },
	["bShowHealthMainDamaged"] = { default=true, nControlType=0, fnCallback="OnSettingHealthChanged" },
	--target components
	["bShowMarkerTarget"] = { default=true, nControlType=1, strControlName="TargetedShowMarker" },
	["bShowNameTarget"] = { default=true, nControlType=1, strControlName="TargetedShowName", fnCallback="OnSettingNameChanged" },
	["bShowRewardsTarget"] = { default=true, nControlType=1, strControlName="TargetedShowRewards"},
	["bShowGuildNameTarget"] = { default=true, nControlType=1, strControlName="TargetedShowGuild", fnCallback="OnSettingTitleChanged" },
	["bShowHealthTarget"] = { default=true, nControlType=1, strControlName="TargetedShowHealth", fnCallback="OnSettingHealthChanged" },
	["bShowRangeTarget"] = { default=false, nControlType=0 },
	["bShowCastBarTargetHostile"] = { default=true, nControlType=1, strControlName="TargetedShowCastBarHostile" },
	["bShowCastBarTargetFriendly"] = { default=false, nControlType=1, strControlName="TargetedShowCastBarFriendly" },
	["bShowUseOcclusion"] = { default=false, nControlType=1, strControlName="MainUseOcclusion" },
	--Non-targeted nameplates in combat
	["bShowForceCastBar"] = { default=false, nControlType=1, strControlName="ShowForceCastBar" },
	["bShowForceThreat"] = { default=false, nControlType=1, strControlName="ShowForceThreat" },
	
	["bShowThreat"] = { default=true, nControlType=1, strControlName="MainShowThreat" },
	["bShowThreatTank"] = { default=false, nControlType=1, strControlName="MainShowThreatTank" },
	
	["bHideInCombat"] = { default=false, nControlType=0 }
}

function Nameplates:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self

	o.arPreloadUnits = {}
	o.bAddonRestoredOrLoaded = false

	o.arWindowPool = {}
	o.arUnit2Nameplate = {}
	o.arWnd2Nameplate = {}

	o.bPlayerInCombat = false
	o.guildDisplayed = nil
	o.guildWarParty = nil
	o.bRedrawRewardIcons = false

    return o
end

function Nameplates:Init()
    Apollo.RegisterAddon(self, true, nil, {"Tooltips", "RewardIcons"})
end

function Nameplates:OnDependencyError(strDependency, strError)
	return true
end

-----------------------------------------------------------------------------------------------
-- Nameplates OnLoad
-----------------------------------------------------------------------------------------------

function Nameplates:OnLoad()
	Apollo.RegisterEventHandler("UnitCreated", 					"OnPreloadUnitCreated", self)

	self.xmlDoc = XmlDoc.CreateFromFile("Nameplates_Remix.xml")
	self.xmlDoc:RegisterCallback("OnDocumentReady", self)
end

function Nameplates:OnPreloadUnitCreated(unitNew)
	self.arPreloadUnits[unitNew:GetId()] = unitNew
end

function Nameplates:OnDocumentReady()
	Apollo.RemoveEventHandler("UnitCreated", self)

	if self.xmlDoc == nil then
		return
	end

	Apollo.RegisterEventHandler("WindowManagementReady", 	"OnWindowManagementReady", self)

	Apollo.RegisterEventHandler("UnitCreated", 					"OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", 				"OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("VarChange_FrameCount", 		"OnFrame", self)

	Apollo.RegisterEventHandler("UnitTextBubbleCreate", 		"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("UnitTextBubblesDestroyed", 	"OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("TargetUnitChanged", 			"OnTargetUnitChanged", self)
	Apollo.RegisterEventHandler("UnitEnteredCombat", 			"OnEnteredCombat", self)
	Apollo.RegisterEventHandler("UnitNameChanged", 				"OnUnitNameChanged", self)
	Apollo.RegisterEventHandler("UnitTitleChanged", 			"OnUnitTitleChanged", self)
	Apollo.RegisterEventHandler("PlayerTitleChange", 			"OnPlayerTitleChanged", self)
	Apollo.RegisterEventHandler("UnitGuildNameplateChanged", 	"OnUnitGuildNameplateChanged",self)
	Apollo.RegisterEventHandler("UnitLevelChanged", 			"OnUnitLevelChanged", self)
	Apollo.RegisterEventHandler("UnitMemberOfGuildChange", 		"OnUnitMemberOfGuildChange", self)
	Apollo.RegisterEventHandler("GuildChange", 					"OnGuildChange", self)
	Apollo.RegisterEventHandler("UnitGibbed",					"OnUnitGibbed", self)
	Apollo.RegisterEventHandler("SpecChanged", 					"SpecChanged", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", 					"Firsttime", self)

	local tRewardUpdateEvents = {
		"QuestObjectiveUpdated", "QuestStateChanged", "ChallengeAbandon", "ChallengeLeftArea",
		"ChallengeFailTime", "ChallengeFailArea", "ChallengeActivate", "ChallengeCompleted",
		"ChallengeFailGeneric", "PublicEventObjectiveUpdate", "PublicEventUnitUpdate",
		"PlayerPathMissionUpdate", "FriendshipAdd", "FriendshipPostRemove", "FriendshipUpdate"
	}

	for i, str in pairs(tRewardUpdateEvents) do
		Apollo.RegisterEventHandler(str, "RequestUpdateAllNameplateRewards", self)
	end

	self.wndMain = Apollo.LoadForm(self.xmlDoc, "NameplatesForm", nil, self)

	Apollo.RegisterTimerHandler("VisibilityTimer", "OnVisibilityTimer", self)
	Apollo.CreateTimer("VisibilityTimer", 0.5, true)

	self.wndOptionsMain = Apollo.LoadForm(self.xmlDoc, "StandardModule", self.wndMain:FindChild("ContentMain"), self)
	self.wndOptionsTargeted = Apollo.LoadForm(self.xmlDoc, "TargetedModule", self.wndMain:FindChild("ContentTarget"), self)
	self.wndMain:Show(false)
	self.wndMain:FindChild("ContentMain"):Show(true)
	self.wndMain:FindChild("ContentTarget"):Show(false)
	self.wndMain:FindChild("ContentToggleContainer:NormalViewCheck"):SetCheck(true)

	self.arUnit2Nameplate = {}
	self.arWnd2Nameplate = {}

	for property,tData in pairs(karSavedProperties) do
		if self[property] == nil then
			self[property] = tData.default
		end
		if tData.nControlType == 1 then
			local wndControl = self.wndMain:FindChild(tData.strControlName)
			if wndControl ~= nil then
				wndControl:SetData(property)
			end
		end
	end
	self.bUseOcclusion = Apollo.GetConsoleVariable("ui.occludeNameplatePositions")

	for key, guildCurr in pairs(GuildLib.GetGuilds()) do
		local eGuildType = guildCurr:GetType()
		if eGuildType == GuildLib.GuildType_Guild then
			self.guildDisplayed = guildCurr
		end
		if eGuildType == GuildLib.GuildType_WarParty then
			self.guildWarParty = guildCurr
		end
	end

	-- Cache defaults
	local wndTemp = Apollo.LoadForm(self.xmlDoc, "NameplateNew", nil, self)
	self.nFrameLeft, self.nFrameTop, self.nFrameRight, self.nFrameBottom = wndTemp:FindChild("Container:HealthBars:MaxHealth"):GetAnchorOffsets()
	self.nHealthWidth = self.nFrameRight - self.nFrameLeft
	self.nFrameLeft2, self.nFrameTop2, self.nFrameRight2, self.nFrameBottom2 = wndTemp:FindChild("Container:ShieldBars:MaxShield"):GetAnchorOffsets()
	self.nShieldBarsWidth = self.nFrameRight2 - self.nFrameLeft2
	
	wndTemp:Destroy()

	self:CreateUnitsFromPreload()
end

function Nameplates:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	local tSave = {}
	for property,tData in pairs(karSavedProperties) do
		tSave[property] = self[property]
	end
	Allsave[currentspec] = tSave
	Allsave["currentspec"] = currentspec
	return Allsave
end

function Nameplates:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	currentspec = tSavedData["currentspec"]
	
	if currentspec ~= nil then
		Allsave = tSavedData
		
		for property,tData in pairs(karSavedProperties) do
			if tSavedData[currentspec][property] ~= nil then
				self[property] = tSavedData[currentspec][property]
			end
		end
	end
	
	self:CreateUnitsFromPreload()
end

function Nameplates: SpecChanged()
	if Allsave["currentspec"] == nil then
		currentspec = AbilityBook.GetCurrentSpec()
		Allsave["currentspec"] = currentspec
	end
	local tSave = {}
	for property,tData in pairs(karSavedProperties) do
		tSave[property] = self[property]
	end
	Allsave[currentspec] = tSave
	currentspec = AbilityBook.GetCurrentSpec()
	for property,tData in pairs(karSavedProperties) do
		if Allsave[currentspec] ~= nil and Allsave[currentspec][property] ~= nil then
			self[property] = Allsave[currentspec][property]
		end
	end
	self:OnSettingHealthChanged()
	self:OnSettingTitleChanged()
	self:OnSettingNameChanged()
	self:ForceOnOcclusion()
	self:RefreshNameplatesConfigure()
end

function Nameplates:OnWindowManagementReady()
	Event_FireGenericEvent("WindowManagementAdd", {wnd = self.wndMain, strName = Apollo.GetString("Nameplates_Options")})
end

function Nameplates:CreateUnitsFromPreload()
	if self.bAddonRestoredOrLoaded then
		self.unitPlayer = GameLib.GetPlayerUnit()

		-- Process units created while form was loading
		for idUnit, unitNew in pairs(self.arPreloadUnits) do
			self:OnUnitCreated(unitNew)
		end
		self.arPreloadUnits = nil
	end
	self.bAddonRestoredOrLoaded = true
end

function Nameplates:OnVisibilityTimer()
	self:UpdateAllNameplateVisibility()
end

function Nameplates:RequestUpdateAllNameplateRewards()
	self.bRedrawRewardIcons = true
end

function Nameplates:UpdateNameplateRewardInfo(tNameplate)
	local tFlags =
	{
		bVert = false,
		bHideQuests = not self.bShowRewardTypeQuest,
		bHideChallenges = not self.bShowRewardTypeChallenge,
		bHideMissions = not self.bShowRewardTypeMission,
		bHidePublicEvents = not self.bShowRewardTypePublicEvent,
		bHideRivals = not self.bShowRivals,
		bHideFriends = not self.bShowFriends
	}

	if RewardIcons ~= nil and RewardIcons.GetUnitRewardIconsForm ~= nil then
		RewardIcons.GetUnitRewardIconsForm(tNameplate.wnd.questRewards, tNameplate.unitOwner, tFlags)
	end
end

function Nameplates:UpdateAllNameplateVisibility()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		self:UpdateNameplateVisibility(tNameplate)
		if self.bRedrawRewardIcons then
			self:UpdateNameplateRewardInfo(tNameplate)
		end
	end

	self.bRedrawRewardIcons = false
end

function Nameplates:UpdateNameplateVisibility(tNameplate)
	local unitOwner = tNameplate.unitOwner
	local wndNameplate = tNameplate.wndNameplate

	tNameplate.bOnScreen = wndNameplate:IsOnScreen()
	tNameplate.bOccluded = wndNameplate:IsOccluded()
	tNameplate.eDisposition = unitOwner:GetDispositionTo(self.unitPlayer)
	local bNewShow = self:HelperVerifyVisibilityOptions(tNameplate) and self:CheckDrawDistance(tNameplate)
	if bNewShow and not tNameplate.bOccluded then
		local NOutsideH = unitOwner:GetOverheadAnchor()
		local NOutside = 1
		if NOutsideH ~= nil and NOutsideH.y < 50 then NOutside = 0 end
		wndNameplate:SetUnit(unitOwner, NOutside)
	end
	--bug bulle
	if bNewShow ~= tNameplate.bShow then
		wndNameplate:Show(bNewShow)
		tNameplate.bShow = bNewShow
	end
end

function Nameplates:OnUnitCreated(unitNew) -- build main options here
	if unitNew == nil
		or not unitNew:IsValid()
		or not unitNew:ShouldShowNamePlate()
		or unitNew:GetType() == "Collectible"
		or unitNew:GetType() == "PinataLoot" then
		-- Never have nameplates
		return
	end

	local idUnit = unitNew:GetId()
	if self.arUnit2Nameplate[idUnit] ~= nil and self.arUnit2Nameplate[idUnit].wndNameplate:IsValid() then
		return
	end

	local wnd = nil
	local wndReferences = nil
	if next(self.arWindowPool) ~= nil then
		local poolEntry = table.remove(self.arWindowPool)
		wnd = poolEntry[1]
		wndReferences = poolEntry[2]
	end

	if wnd == nil or not wnd:IsValid() then
		wnd = Apollo.LoadForm(self.xmlDoc, "NameplateNew", "InWorldHudStratum", self)
		wndReferences = nil
	end

	wnd:Show(false, true)
	wnd:SetUnit(unitNew, 1)

	local tNameplate =
	{
		unitOwner 		= unitNew,
		idUnit 			= idUnit,
		wndNameplate	= wnd,
		bOnScreen 		= wnd:IsOnScreen(),
		bOccluded 		= wnd:IsOccluded(),
		bSpeechBubble 	= false,
		bIsTarget 		= false,
		bIsCluster 		= false,
		bIsCasting 		= false,
		bGibbed			= false,
		bIsGuildMember 	= self.guildDisplayed and self.guildDisplayed:IsUnitMember(unitNew) or false,
		bIsWarPartyMember = self.guildWarParty and self.guildWarParty:IsUnitMember(unitNew) or false,
		nVulnerableTime = 0,
		eDisposition	= unitNew:GetDispositionTo(self.unitPlayer),
		bShow			= false,
		wnd				= wndReferences,
	}

	if wndReferences == nil then
		tNameplate.wnd =
		{
			health = wnd:FindChild("Container:HealthBars"),
			Shield = wnd:FindChild("Container:ShieldBars"),
			castBar = wnd:FindChild("Container:CastContainer"),
			castBarCastFill = wnd:FindChild("Container:CastContainer:CastFill"),
			vulnerable = wnd:FindChild("Container:Vulnerable"),
			level = wnd:FindChild("NameContainer:Level"),
			armor = wnd:FindChild("NameContainer:Armor"),
			armorLabel = wnd:FindChild("NameContainer:Armor:LabelArmor"),
			wndGuild = wnd:FindChild("Guild"),
			wndName = wnd:FindChild("NameContainer:Name"),
			healthMaxShield = wnd:FindChild("Container:ShieldBars:MaxShield"),
			healthShieldFill = wnd:FindChild("Container:ShieldBars:MaxShield:ShieldFill"),
			healthMaxAbsorb = wnd:FindChild("Container:ShieldBars:MaxAbsorb"),
			healthAbsorbFill = wnd:FindChild("Container:ShieldBars:MaxAbsorb:AbsorbFill"),
			healthMaxHealth = wnd:FindChild("Container:HealthBars:MaxHealth"),
			healthHealthLabel = wnd:FindChild("Container:HealthLabel"),
			ShieldLabel  = wnd:FindChild("Container:ShieldLabel"),
			vulnerableVulnFill = wnd:FindChild("Container:Vulnerable:VulnFill"),
			questRewards = wnd:FindChild("QuestRewards"),
			targetMarker = wnd:FindChild("TDContainer:TargetMarker"),
			debuffMarker = wnd:FindChild("TDContainer:Debuff"),
		}
	end

	self.arUnit2Nameplate[idUnit] = tNameplate
	self.arWnd2Nameplate[wnd:GetId()] = tNameplate

	self:DrawNameAndGuild(tNameplate)
	self:DrawLevel(tNameplate)
	-- drawrewards bug
	self:DrawRewards(tNameplate)
	self:UpdateNameplateRewardInfo(tNameplate)
	
end

function Nameplates:OnUnitDestroyed(unitOwner)
	local idUnit = unitOwner:GetId()
	if self.arUnit2Nameplate[idUnit] == nil then
		return
	end

	local tNameplate = self.arUnit2Nameplate[idUnit]
	local wndNameplate = tNameplate.wndNameplate

	self.arWnd2Nameplate[wndNameplate:GetId()] = nil
	wndNameplate:Destroy()
	self.arUnit2Nameplate[idUnit] = nil
end

function Nameplates:OnFrame()
	self.unitPlayer = GameLib.GetPlayerUnit()

	local unitOwner
	local wndNameplate
	local nCon
	local unitWindow

	local fnDrawHealth = Nameplates.DrawHealth
	local fnHelperCalculateConValue = Nameplates.HelperCalculateConValue
	local fnDrawRewards = Nameplates.DrawRewards
	local fnDrawCastBar = Nameplates.DrawCastBar
	local fnDrawVulnerable = Nameplates.DrawVulnerable
	local fnDrawTargeting = Nameplates.DrawTargeting
	
	local fnColorNameplate = Nameplates.ColorNameplate

	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		if tNameplate.bShow then
			unitOwner = tNameplate.unitOwner
			wndNameplate = tNameplate.wndNameplate
			unitWindow = wndNameplate:GetUnit()

			fnDrawHealth(self, tNameplate)

			nCon = fnHelperCalculateConValue(self, unitOwner)

			fnDrawRewards(self, tNameplate)
			fnDrawCastBar(self, tNameplate)
			fnDrawVulnerable(self, tNameplate)
			fnDrawTargeting(self, tNameplate)
			fnColorNameplate(self, tNameplate)
		end
	end
end

function Nameplates:ColorNameplate(tNameplate)
	local unitPlayer = self.unitPlayer
	local unitOwner = tNameplate.unitOwner
	local wndNameplate = tNameplate.wndNameplate

	local eDisposition = tNameplate.eDisposition
	local nCon = self:HelperCalculateConValue(unitOwner)

	local crLevelColorToUse = karConColors[nCon][2]
	if unitOwner:IsScaled() then
		crLevelColorToUse = CColor.new(1.0, 191/255, 128/255, 0.7)
	elseif unitOwner:GetLevel() == nil then
		crLevelColorToUse = karConColors[1][2]
	end

	local crColorToUse = karDisposition.tTextColors[eDisposition]
	local unitController = unitOwner:GetUnitOwner() or unitOwner
	local strUnitType = unitOwner:GetType()

	if strUnitType == "Player" or strUnitType == "Pet" or strUnitType == "Esper Pet" then
		if eDisposition == Unit.CodeEnumDisposition.Friendly or unitOwner:IsThePlayer() then
			crColorToUse = kcrUnflaggedAlly
			if unitController:IsInYourGroup() then
				crColorToUse = kcrUnflaggedGroupmate
			elseif tNameplate.bIsGuildMember then
				crColorToUse = kcrUnflaggedGuildmate
			elseif unitController:IsPvpFlagged() then
				crColorToUse = kcrFlaggedAlly
			end
		else
			local bIsUnitFlagged = unitController:IsPvpFlagged()
			local bAmIFlagged = GameLib.IsPvpFlagged()

			if not bAmIFlagged and not bIsUnitFlagged then
				crColorToUse = kcrUnflaggedEnemyWhenUnflagged
			elseif bAmIFlagged and not bIsUnitFlagged then
				crColorToUse = kcrUnflaggedEnemyWhenFlagged
			elseif not bAmIFlagged and bIsUnitFlagged then
				crColorToUse = kcrFlaggedEnemyWhenUnflagged
			elseif bAmIFlagged and bIsUnitFlagged then
				crColorToUse = kcrFlaggedEnemyWhenFlagged
			end
		end
	end

	if unitOwner:GetType() ~= "Player" and unitOwner:IsTagged() and not unitOwner:IsTaggedByMe() and not unitOwner:IsSoftKill() then
		crColorToUse = kcrDefaultTaggedColor
	end

	if unitOwner:IsDead() then
		crColorToUse = kcrDeadColor
		crLevelColorToUse = kcrDeadColor
	end

	tNameplate.wnd.level:SetTextColor(crLevelColorToUse)
	tNameplate.wnd.wndName:SetTextColor(crColorToUse)
	tNameplate.wnd.wndGuild:SetTextColor(crColorToUse)
end

function Nameplates:DrawNameAndGuild(tNameplate)

	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner

	local wndName = tNameplate.wnd.wndName
	local bShowName = self.bShowNameMain
	
	local wndGuild = tNameplate.wnd.wndGuild
	local bShowGuild = self.bShowTitle
	local updateGuild = false
	local strNewGuild = ""
	
	local bUseTarget = tNameplate.bIsTarget
	
	if bUseTarget then
		bShowGuild = self.bShowGuildNameTarget
		bShowName = self.bShowNameTarget
	end
	
	if unitOwner:GetType() == "Player" then
		strNewGuild = unitOwner:GetAffiliationName()
		if strNewGuild ~= nil and strNewGuild ~= "" then
			strNewGuild = String_GetWeaselString(Apollo.GetString("Nameplates_GuildDisplay"), strNewGuild)

		end
	end

	if bShowGuild and strNewGuild ~= wndGuild:GetText() then
		wndGuild:SetTextRaw(strNewGuild)
		updateGuild = true
	end

	wndGuild:Show(bShowGuild and strNewGuild ~= nil and strNewGuild ~= "")
	
	if wndName:IsShown() ~= bShowName then
		wndName:Show(bShowName)
	end
	
	if bShowName or bShowGuild then
		local strNewName
		if self.bShowTitle then
			strNewName = unitOwner:GetTitleOrName()
		else
			strNewName = unitOwner:GetName()
		end

		if wndName:GetText() ~= strNewName or updateGuild then
			wndName:SetText(strNewName)

			-- Resize
			local wndNameplate = tNameplate.wndNameplate
			local nLeft, nTop, nRight, nBottom = wndNameplate:GetAnchorOffsets()
			local nHalfNameWidth = math.ceil(math.max(Apollo.GetTextWidth("Nameplates", strNewName)+50, Apollo.GetTextWidth("CRB_Interface9_BO", strNewGuild)) / 2)
			nHalfNameWidth = math.max(nHalfNameWidth, self.nHealthWidth / 2)
			wndNameplate:SetAnchorOffsets(-nHalfNameWidth, nTop, nHalfNameWidth, nBottom)
		end
	end
	wndNameplate:ArrangeChildrenVert(2) -- Must be run if bShowGuild is false as well
end

function Nameplates:DrawLevel(tNameplate)
	local unitOwner = tNameplate.unitOwner

	if self.bShowLevel then
		tNameplate.wnd.level:SetText(unitOwner:GetLevel() or "")
	else
		tNameplate.wnd.level:SetText("")
	end
end

function Nameplates:DrawHealth(tNameplate)

	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner

	local wndHealth = tNameplate.wnd.health
	local wndShield = tNameplate.wnd.Shield

	if unitOwner:GetHealth() == nil or unitOwner:GetHealth() == 0 then
		wndHealth:Show(false)
		wndShield:Show(false)
		tNameplate.wnd.healthHealthLabel:SetText("")
		tNameplate.wnd.ShieldLabel:SetText("")
		return
	end

	local bUseTarget = tNameplate.bIsTarget
	if bUseTarget then
		wndHealth:Show(self.bShowHealthTarget)
		wndShield:Show(self.bShowHealthTarget)
	else
		if self.bShowHealthMain then
			wndHealth:Show(true)
			wndShield:Show(true)
		elseif self.bShowHealthMainDamaged then
			wndHealth:Show(unitOwner:GetHealth() ~= unitOwner:GetMaxHealth())
			if unitOwner:IsInYourGroup() then
				wndShield:Show(unitOwner:GetShieldCapacity() < (unitOwner:GetShieldCapacityMax()/1.5))
				
			else
				wndShield:Show(unitOwner:GetShieldCapacity() < unitOwner:GetShieldCapacityMax())
			end
		else
			wndHealth:Show(false)
			wndShield:Show(false)
		end
	end

	if wndHealth:IsShown() or wndShield:IsShown() then
		self:HelperDoHealthShieldBar(wndHealth, unitOwner, tNameplate.eDisposition, tNameplate)
	else
	
		tNameplate.wnd.healthHealthLabel:SetText("")
		tNameplate.wnd.ShieldLabel:SetText("")
		self:SDebuffs(unitOwner, tNameplate)
		self:SArmor(unitOwner, tNameplate)
		tNameplate.wnd.Shield:Show(false)
		tNameplate.wnd.healthShieldFill:Show(false)
		tNameplate.wnd.healthMaxHealth:Show(false)
		tNameplate.wnd.healthMaxShield:Show(false)
		tNameplate.wnd.healthMaxAbsorb:Show(false)
		tNameplate.wnd.vulnerable:Show(false)
	end
end

function Nameplates:DrawCastBar(tNameplate)
	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner

	-- Casting; has some onDraw parameters we need to check
	tNameplate.bIsCasting = unitOwner:ShouldShowCastBar()
	local eDisposition = tNameplate.eDisposition
	local bShowTarget = tNameplate.bIsTarget
	local wndCastBar = tNameplate.wnd.castBar
		
	if tNameplate.bIsCasting and bShowTarget then
		if self.bShowCastBarTargetHostile and eDisposition == Unit.CodeEnumDisposition.Hostile then
			bShow = true
		elseif self.bShowCastBarTargetHostile and eDisposition == Unit.CodeEnumDisposition.Neutral then
			bShow = true
		elseif self.bShowCastBarTargetFriendly and eDisposition == Unit.CodeEnumDisposition.Friendly then
			bShow = true
		else bShow = false
		end
	else
		if self.bShowCastBarMainHostile and eDisposition == Unit.CodeEnumDisposition.Hostile then
			bShow = tNameplate.bIsCasting
		elseif self.bShowCastBarMainHostile and eDisposition == Unit.CodeEnumDisposition.Neutral then
			bShow = tNameplate.bIsCasting
		elseif self.bShowCastBarMainFriendly and eDisposition == Unit.CodeEnumDisposition.Friendly then
			bShow = tNameplate.bIsCasting
		else bShow = false
		end
	end

	wndCastBar:Show(bShow)
	if bShow then
		tNameplate.wnd.castBar:SetText(" "..unitOwner:GetCastName())
		tNameplate.wnd.castBarCastFill:SetMax(unitOwner:GetCastDuration())
		tNameplate.wnd.castBarCastFill:SetProgress(unitOwner:GetCastElapsed())
	end
end

function Nameplates:DrawVulnerable(tNameplate)
	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner

	local bUseTarget = tNameplate.bIsTarget
	local wndVulnerable = tNameplate.wnd.vulnerable

	local bIsVulnerable = false
	if (not bUseTarget and (self.bShowHealthMain or self.bShowHealthMainDamaged)) or (bUseTarget and self.bShowHealthTarget) then
		local nVulnerable = unitOwner:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability)
		if nVulnerable == nil then
			wndVulnerable:Show(false)
		elseif nVulnerable == 0 and nVulnerable ~= tNameplate.nVulnerableTime then
			tNameplate.nVulnerableTime = 0 -- casting done, set back to 0
			wndVulnerable:Show(false)
		elseif nVulnerable ~= 0 and nVulnerable > tNameplate.nVulnerableTime then
			tNameplate.nVulnerableTime = nVulnerable
			wndVulnerable:Show(true)
			bIsVulnerable = true
		elseif nVulnerable ~= 0 and nVulnerable < tNameplate.nVulnerableTime then
			tNameplate.wnd.vulnerableVulnFill:SetMax(tNameplate.nVulnerableTime)
			tNameplate.wnd.vulnerableVulnFill:SetProgress(nVulnerable)
			bIsVulnerable = true
		end
	end
end

function Nameplates:DrawRewards(tNameplate)
	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner

	local bUseTarget = tNameplate.bIsTarget
	local bShow = self.bShowRewardsMain
	if bUseTarget then
		bShow = self.bShowRewardsTarget
	end

	tNameplate.wnd.questRewards:Show(bShow)

end

function Nameplates:DrawTargeting(tNameplate)
	local wndNameplate = tNameplate.wndNameplate
	local unitOwner = tNameplate.unitOwner
	local dead = false
	local danger = 0
	local bUseTarget = tNameplate.bIsTarget
	local bShowTargetMarker = bUseTarget and self.bShowMarkerTarget and tNameplate.wnd.health:IsShown()
	local eDisposition = tNameplate.eDisposition
	local UTarget = unitOwner:GetTarget() 	
	local unitPlayer = self.unitPlayer
	
	if self.bshowHealIcons and unitOwner:IsInYourGroup() then
		
		if bShowTargetMarker then
			tNameplate.wnd.targetMarker:SetSprite("ClientSprites:PlusHealthBarFull")
		else
			tNameplate.wnd.targetMarker:SetSprite("CRB_MinimapSprites:sprMM_QuestArrow")
			tNameplate.wnd.targetMarker:SetRotation(180)
		end
		local nHealthMax = unitOwner:GetMaxHealth()
		local nHealth = unitOwner:GetHealth()
		if unitOwner ~= nil and nHealthMax ~= nil and nHealth ~= nil  then
			if unitOwner:IsDead() then
				dead = true
			else
				danger = math.floor(100-(100 / nHealthMax * nHealth))
			end
			if not unitOwner:IsThePlayer() and danger > 7 then
				bShowTargetMarker = true
			end
		end

		if dead == true then
			tNameplate.wnd.targetMarker:SetBGColor("black")
		elseif danger >= 74 then
			tNameplate.wnd.targetMarker:SetBGColor("red")
		elseif danger >= 49 then
			tNameplate.wnd.targetMarker:SetBGColor("xkcdOrange")
		elseif danger >= 24 then
			tNameplate.wnd.targetMarker:SetBGColor("Yellow")
		elseif danger >= 9 then
			tNameplate.wnd.targetMarker:SetBGColor("blue")
		else
			tNameplate.wnd.targetMarker:SetBGColor("Green")
		end
	elseif UTarget ~= nil and eDisposition ~= Unit.CodeEnumDisposition.Friendly and (self.bShowThreatTank or self.bShowThreat) and self.unitPlayer:IsInYourGroup() then
		UTargetPlayer = UTarget:IsThePlayer()
		if bShowTargetMarker then
			if UTargetPlayer then
				tNameplate.wnd.targetMarker:SetSprite("ClientSprites:ComboStarFull")
				if self.bShowThreatTank then
					tNameplate.wnd.targetMarker:SetBGColor("green")
				else
					tNameplate.wnd.targetMarker:SetBGColor("red")
				end
			elseif self.bShowThreatTank and UTarget:IsInYourGroup() then
				tNameplate.wnd.targetMarker:SetSprite("ClientSprites:ComboStarFull")
				tNameplate.wnd.targetMarker:SetBGColor("red")
			else
				tNameplate.wnd.targetMarker:SetSprite("zonemap:UI_WorldMap_Position")
				tNameplate.wnd.targetMarker:SetBGColor(karDisposition.tTargetPrimary[tNameplate.eDisposition])
			end	
			tNameplate.wnd.targetMarker:SetRotation(0)
		else
			if UTargetPlayer then
				tNameplate.wnd.targetMarker:SetSprite("ClientSprites:ComboStarEmpty")
				tNameplate.wnd.targetMarker:SetRotation(0)		
				bShowTargetMarker = true
				if self.bShowThreatTank then
					tNameplate.wnd.targetMarker:SetBGColor("green")
				else
					tNameplate.wnd.targetMarker:SetBGColor("red")
				end
			elseif self.bShowThreatTank and UTarget:IsInYourGroup() then
				bShowTargetMarker = true
				tNameplate.wnd.targetMarker:SetSprite("ClientSprites:ComboStarEmpty")
				tNameplate.wnd.targetMarker:SetRotation(0)		
				tNameplate.wnd.targetMarker:SetBGColor("red")
			end			
		end
	else
		tNameplate.wnd.targetMarker:SetSprite("zonemap:UI_WorldMap_Position")
		tNameplate.wnd.targetMarker:SetBGColor(karDisposition.tTargetPrimary[tNameplate.eDisposition])
		tNameplate.wnd.targetMarker:SetRotation(0)
	end
	
	if tNameplate.wnd.targetMarker:IsShown() ~= bShowTargetMarker then
		tNameplate.wnd.targetMarker:Show(bShowTargetMarker)
	end
end

function Nameplates:CheckDrawDistance(tNameplate)
	local unitPlayer = self.unitPlayer
	local unitOwner = tNameplate.unitOwner

	if not unitOwner or not unitPlayer then
	    return false
	end

	local tPosTarget = unitOwner:GetPosition()
	local tPosPlayer = unitPlayer:GetPosition()

	if tPosTarget == nil or tPosPlayer == nil then
		return
	end

	local nDeltaX = tPosTarget.x - tPosPlayer.x
	local nDeltaY = tPosTarget.y - tPosPlayer.y
	local nDeltaZ = tPosTarget.z - tPosPlayer.z

	local nDistance = (nDeltaX * nDeltaX) + (nDeltaY * nDeltaY) + (nDeltaZ * nDeltaZ)

	if tNameplate.bIsTarget or tNameplate.bIsCluster then
		bInRange = nDistance < knTargetRange
		return bInRange
	else
		bInRange = nDistance < (self.nMaxRange * self.nMaxRange) -- squaring for quick maths
		return bInRange
	end
end

function Nameplates:HelperVerifyVisibilityOptions(tNameplate)
	local unitPlayer = self.unitPlayer
	local unitOwner = tNameplate.unitOwner
	local eDisposition = tNameplate.eDisposition
	local UInGroup = unitOwner:IsInYourGroup()

	local bHiddenUnit = not unitOwner:ShouldShowNamePlate()
	if bHiddenUnit and not tNameplate.bIsTarget then
		return false
	end

	if (self.bUseOcclusion and tNameplate.bOccluded) or not tNameplate.bOnScreen then
		return false
	end

	local bShowNameplate = false

	if self.bShowMainObjectiveOnly and tNameplate.bIsObjective then
		bShowNameplate = true
	end

	if self.bShowMainGroupOnly and UInGroup then
		bShowNameplate = true
	end

	if self.bShowDispositionHostile and eDisposition == Unit.CodeEnumDisposition.Hostile then
		bShowNameplate = true
	end

	if self.bShowDispositionNeutral and eDisposition == Unit.CodeEnumDisposition.Neutral then
		bShowNameplate = true
	end

	if self.bShowDispositionFriendly and eDisposition == Unit.CodeEnumDisposition.Friendly then
		bShowNameplate = true
	end

	if self.bShowDispositionFriendlyPlayer and eDisposition == Unit.CodeEnumDisposition.Friendly and unitOwner:GetType() == "Player" then
		bShowNameplate = true
	end

	local tActivation = unitOwner:GetActivationState()

	if self.bShowVendor and tActivation.Vendor ~= nil then
		bShowNameplate = true
	end

	if self.bShowTaxi and (tActivation.FlightPathSettler ~= nil or tActivation.FlightPath ~= nil or tActivation.FlightPathNew) then
		bShowNameplate = true
	end

	if self.bShowOrganization and tNameplate.bIsGuildMember then
		bShowNameplate = true
	end

	if self.bShowMainObjectiveOnly then
		-- QuestGivers too
		if tActivation.QuestReward ~= nil then
			bShowNameplate = true
		end

		if tActivation.QuestNew ~= nil or tActivation.QuestNewMain ~= nil then
			bShowNameplate = true
		end

		if tActivation.QuestReceiving ~= nil then
			bShowNameplate = true
		end

		if tActivation.TalkTo ~= nil then
			bShowNameplate = true
		end
	end

	if bShowNameplate and not UInGroup then
		bShowNameplate = not (self.bPlayerInCombat and self.bHideInCombat)
	end
	
	if not bShowNameplate and  self.bShowForceCastBar and eDisposition ~= Unit.CodeEnumDisposition.Friendly then
		tNameplate.bIsCasting = unitOwner:ShouldShowCastBar()
		bShowNameplate = tNameplate.bIsCasting
	end
	
	if not bShowNameplate and  self.bShowForceThreat and (self.bShowThreatTank or self.bShowThreat) and tNameplate.wnd.targetMarker:IsShown() then
		bShowNameplate = true
	end
	
	if unitOwner:IsThePlayer() then
		if self.bShowMyNameplate and not unitOwner:IsDead() then
			bShowNameplate = true
		else
			bShowNameplate = false
		end
	end
	
	 -- bug bulle
	if self.bPlayerInCombat and bShowNameplate and (tNameplate.bGibbed or tNameplate.bSpeechBubble) then
		bShowNameplate = true
	elseif tNameplate.bGibbed or tNameplate.bSpeechBubble then
		bShowNameplate = false
	end

	return bShowNameplate or tNameplate.bIsTarget
end

function Nameplates:HelperDoHealthShieldBar(wndHealth, unitOwner, eDisposition, tNameplate)
	local nVulnerabilityTime = unitOwner:GetCCStateTimeRemaining(Unit.CodeEnumCCState.Vulnerability)

	if unitOwner:GetType() == "Simple" or unitOwner:GetHealth() == nil then
		tNameplate.wnd.healthMaxHealth:SetAnchorOffsets(self.nFrameLeft, self.nFrameTop, self.nFrameRight, self.nFrameBottom)
		tNameplate.wnd.healthHealthLabel:SetText("")
		return
	end

	local nHealthCurr 	= unitOwner:GetHealth()
	local nHealthMax 	= unitOwner:GetMaxHealth()
	local nShieldCurr 	= unitOwner:GetShieldCapacity()
	local nShieldMax 	= unitOwner:GetShieldCapacityMax()
	local nAbsorbCurr 	= 0
	local nAbsorbMax 	= unitOwner:GetAbsorptionMax()
	if nAbsorbMax > 0 then
		nAbsorbCurr = unitOwner:GetAbsorptionValue() -- Since it doesn't clear when the buff drops off
	end

	local nTotalMax = nShieldMax + nAbsorbMax

	if unitOwner:IsDead() then
		nHealthCurr = 0
	end

	local nPointHealthRight = self.nFrameLeft + (self.nHealthWidth * (nHealthCurr / nHealthMax)) -- applied to the difference between L and R
	local nPointShieldRight = self.nFrameLeft2 + (self.nShieldBarsWidth * (nShieldMax / nTotalMax))
	local nPointAbsorbRight = self.nFrameLeft2 + (self.nShieldBarsWidth * ((nShieldMax + nAbsorbMax) / nTotalMax))

	-- Resize
	self:SetBarValue(tNameplate.wnd.healthShieldFill, 0, nShieldCurr, nShieldMax) -- Only the Curr Shield really progress fills
	self:SetBarValue(tNameplate.wnd.healthAbsorbFill, 0, nAbsorbCurr, nAbsorbMax)
	tNameplate.wnd.healthMaxHealth:SetAnchorOffsets(self.nFrameLeft, self.nFrameTop, nPointHealthRight, self.nFrameBottom)
	tNameplate.wnd.healthMaxShield:SetAnchorOffsets(self.nFrameLeft2, self.nFrameTop2, nPointShieldRight, self.nFrameBottom2)
	tNameplate.wnd.healthMaxAbsorb:SetAnchorOffsets(nPointShieldRight, self.nFrameTop2, nPointAbsorbRight, self.nFrameBottom2)

	-- Bars
	tNameplate.wnd.healthShieldFill:Show(nHealthCurr > 0)
	tNameplate.wnd.healthMaxHealth:Show(nHealthCurr > 0)
	tNameplate.wnd.healthMaxShield:Show(nShieldCurr > 0 and nShieldMax > 0)
	tNameplate.wnd.healthMaxAbsorb:Show(nAbsorbCurr > 0 and nAbsorbMax > 0)
	tNameplate.wnd.Shield:Show(nShieldMax > 0)

	-- Text
	local strHealthMax = self:HelperFormatBigNumber(nHealthMax)
	local strHealthCurr = self:HelperFormatBigNumber(nHealthCurr)
	local strShieldCurr = self:HelperFormatBigNumber(nShieldCurr)

	tNameplate.wnd.healthHealthLabel:SetText(strHealthCurr)
	tNameplate.wnd.ShieldLabel:SetText("( "..strShieldCurr.." )")
	
	self:SDebuffs(unitOwner, tNameplate)
	self:SArmor(unitOwner, tNameplate)
	
	-- Sprite
	if (nVulnerabilityTime and nVulnerabilityTime > 0) or tNameplate.nVulnerableTime == -1 then
		tNameplate.wnd.healthMaxHealth:SetBGColor("xkcdAmethyst")
	else
		tNameplate.wnd.healthMaxHealth:SetBGColor(karDisposition.tHealthBar[eDisposition])
	end

end

function Nameplates:SDebuffs(unitOwner, tNameplate)
	-- debuff
	local NoDebuffShow = true
	local UnitGrp = unitOwner:IsInYourGroup()
	if(self.bShowIconsDebuff and UnitGrp) then
		local myBuffs = unitOwner:GetBuffs()
		local debuffs = myBuffs["arHarmful"]
		if debuffs then
			for key, value in pairs(debuffs) do
				if value["splEffect"]:GetClass() == 38 and value["splEffect"]:GetName() ~= "Bond" 
				then
					if UnitGrp then
						tNameplate.wnd.debuffMarker:Show(true)
						NoDebuffShow = false
						tNameplate.nVulnerableTime = -1
					end
					break
				end
			end
		end		
	end
	if NoDebuffShow then
		tNameplate.wnd.debuffMarker:Show(false)
		if tNameplate.nVulnerableTime == -1 then
			tNameplate.nVulnerableTime = 0
		end
	end
		--fin
end

function Nameplates:SArmor(unitOwner, tNameplate)
	local Armornb = unitOwner:GetInterruptArmorValue()
	if Armornb > 0 and (self.bShowNameMain or self.bShowNameTarget) then
		tNameplate.wnd.armorLabel:SetText(Armornb)
		tNameplate.wnd.armor:Show(true)
	else
		tNameplate.wnd.armor:Show(false)
	end
end

function Nameplates:HelperFormatBigNumber(nArg)
	if nArg < 1000 then
		strResult = tostring(nArg)
	elseif nArg < 1000000 then
		if math.floor(nArg%1000/100) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberWhole"), math.floor(nArg / 1000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_ShortNumberFloat"), nArg / 1000)
		end
	elseif nArg < 1000000000 then
		if math.floor(nArg%1000000/100000) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_MillionsNumberWhole"), math.floor(nArg / 1000000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_MillionsNumberFloat"), nArg / 1000000)
		end
	elseif nArg < 1000000000000 then
		if math.floor(nArg%1000000/100000) == 0 then
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_BillionsNumberWhole"), math.floor(nArg / 1000000))
		else
			strResult = String_GetWeaselString(Apollo.GetString("TargetFrame_BillionsNumberFloat"), nArg / 1000000)
		end
	else
		strResult = tostring(nArg)
	end
	return strResult
end

function Nameplates:SetBarValue(wndBar, fMin, fValue, fMax)
	wndBar:SetMax(fMax)
	wndBar:SetFloor(fMin)
	wndBar:SetProgress(fValue)
end

function Nameplates:HelperCalculateConValue(unitTarget)
	if unitTarget == nil or self.unitPlayer == nil then
		return 1
	end

	local nUnitCon = self.unitPlayer:GetLevelDifferential(unitTarget)

	local nCon = 1 --default setting

	if nUnitCon <= karConColors[1][1] then -- lower bound
		nCon = 1
	elseif nUnitCon >= karConColors[#karConColors][1] then -- upper bound
		nCon = #karConColors
	else
		for idx = 2, (#karConColors - 1) do -- everything in between
			if nUnitCon == karConColors[idx][1] then
				nCon = idx
			end
		end
	end

	return nCon
end

-----------------------------------------------------------------------------------------------
-- Nameplate Events
-----------------------------------------------------------------------------------------------

function Nameplates:OnNameplateNameClick(wndHandler, wndCtrl, eMouseButton)
	local tNameplate = self.arWnd2Nameplate[wndHandler:GetId()]
	if tNameplate == nil then
		return
	end

	local unitOwner = tNameplate.unitOwner
	if GameLib.GetTargetUnit() ~= unitOwner and eMouseButton == GameLib.CodeEnumInputMouse.Left then
		GameLib.SetTargetUnit(unitOwner)
	end
end

function Nameplates:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
	local tNameplate = self.arWnd2Nameplate[wndHandler:GetId()]
	if tNameplate ~= nil then
		tNameplate.bOnScreen = bOnScreen
		self:UpdateNameplateVisibility(tNameplate)
	end
end

function Nameplates:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
	local tNameplate = self.arWnd2Nameplate[wndHandler:GetId()]
	if tNameplate ~= nil then
		tNameplate.bOccluded = bOccluded
		self:UpdateNameplateVisibility(tNameplate)
	end
end

-----------------------------------------------------------------------------------------------
-- System Events
-----------------------------------------------------------------------------------------------

function Nameplates:OnUnitTextBubbleToggled(tUnitArg, strText, nRange)
--bug bulle
	local tNameplate = self.arUnit2Nameplate[tUnitArg:GetId()]
	if tNameplate ~= nil then
		tNameplate.bSpeechBubble = strText ~= nil and strText ~= ""
		self:UpdateNameplateVisibility(tNameplate)
	end
end

function Nameplates:OnEnteredCombat(unitChecked, bInCombat)
	if unitChecked == self.unitPlayer then
		self.bPlayerInCombat = bInCombat
	end
end

function Nameplates:OnUnitGibbed(unitUpdated)
--bug?
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate ~= nil then
		tNameplate.bGibbed = true
		self:UpdateNameplateVisibility(tNameplate)
	end
end

function Nameplates:OnUnitNameChanged(unitUpdated, strNewName)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate ~= nil then
		self:DrawNameAndGuild(tNameplate)
	end
end

function Nameplates:OnUnitTitleChanged(unitUpdated)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate ~= nil then
		self:DrawNameAndGuild(tNameplate)
	end
end

function Nameplates:OnPlayerTitleChanged()
	local tNameplate = self.arUnit2Nameplate[self.unitPlayer:GetId()]
	if tNameplate ~= nil then
		self:DrawNameAndGuild(tNameplate)
	end
end

function Nameplates:OnUnitLevelChanged(unitUpdating)
	local tNameplate = self.arUnit2Nameplate[unitUpdating:GetId()]
	if tNameplate ~= nil then
		self:DrawLevel(tNameplate)
	end
end

function Nameplates:OnGuildChange()
	self.guildDisplayed = nil
	self.guildWarParty = nil
	for key, guildCurr in pairs(GuildLib.GetGuilds()) do
		local eGuildType = guildCurr:GetType()
		if eGuildType == GuildLib.GuildType_Guild then
			self.guildDisplayed = guildCurr
		end
		if eGuildType == GuildLib.GuildType_WarParty then
			self.guildWarParty = guildCurr
		end
	end

	for key, tNameplate in pairs(self.arUnit2Nameplate) do
		local unitOwner = tNameplate.unitOwner
		tNameplate.bIsGuildMember = self.guildDisplayed and self.guildDisplayed:IsUnitMember(unitOwner) or false
		tNameplate.bIsWarPartyMember = self.guildWarParty and self.guildWarParty:IsUnitMember(unitOwner) or false
	end
end

function Nameplates:OnUnitGuildNameplateChanged(unitUpdated)
	local tNameplate = self.arUnit2Nameplate[unitUpdated:GetId()]
	if tNameplate ~= nil then
		self:DrawNameAndGuild(tNameplate)
	end
end

function Nameplates:OnUnitMemberOfGuildChange(unitOwner)
	local tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
	if tNameplate ~= nil then
		self:DrawNameAndGuild(tNameplate)
		tNameplate.bIsGuildMember = self.guildDisplayed and self.guildDisplayed:IsUnitMember(unitOwner) or false
		tNameplate.bIsWarPartyMember = self.guildWarParty and self.guildWarParty:IsUnitMember(unitOwner) or false
	end
end

function Nameplates:OnTargetUnitChanged(unitOwner) -- build targeted options here; we get this event when a creature attacks, too

	for idx, tNameplateOther in pairs(self.arUnit2Nameplate) do
		local bIsTarget = tNameplateOther.bIsTarget
		local bIsCluster = tNameplateOther.bIsCluster

		tNameplateOther.bIsTarget = false
		tNameplateOther.bIsCluster = false

		if bIsTarget or bIsCluster then
			self:DrawNameAndGuild(tNameplateOther)
			self:DrawLevel(tNameplateOther)
			self:UpdateNameplateRewardInfo(tNameplateOther)
		end
	end

	if unitOwner == nil then
		return
	end

	local tNameplate = self.arUnit2Nameplate[unitOwner:GetId()]
	if tNameplate == nil then
		return
	end

	if GameLib.GetTargetUnit() == unitOwner then
		tNameplate.bIsTarget = true
		self:DrawNameAndGuild(tNameplate)
		self:DrawLevel(tNameplate)
		self:UpdateNameplateRewardInfo(tNameplate)

		local tCluster = unitOwner:GetClusterUnits()
		if tCluster ~= nil then
			tNameplate.bIsCluster = true

			for idx, unitCluster in pairs(tCluster) do
				local tNameplateOther = self.arUnit2Nameplate[unitCluster:GetId()]
				if tNameplateOther ~= nil then
					tNameplateOther.bIsCluster = true
				end
			end
		end
	end
end

-----------------------------------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------------------------------
function Nameplates:OnConfigure()
	self:OnNameplatesOn()
end

function Nameplates:OnNameplatesOn()
	local ePath = PlayerPathLib.GetPlayerPathType()
	self.wndOptionsMain:FindChild("ShowRewardTypeMission"):FindChild("Icon"):SetSprite(karPathSprite[ePath])
	self.wndMain:Show(true)
	self:RefreshNameplatesConfigure()
end

function Nameplates:RefreshNameplatesConfigure()
	-- Generic managed controls
	for property,tData in pairs(karSavedProperties) do
		if tData.nControlType == 1 and self[property] ~= nil then
			local wndControl = self.wndMain:FindChild(tData.strControlName)
			if wndControl ~= nil then
				wndControl:SetCheck(self[property])
			end
		end
	end

	--Draw distance
	if self.nMaxRange ~= nil then
		self.wndOptionsMain:FindChild("ShowOptionsBacker:DrawDistanceSlider"):SetValue(self.nMaxRange)
		self.wndOptionsMain:FindChild("ShowOptionsBacker:DrawDistanceLabel"):SetText(String_GetWeaselString(Apollo.GetString("Nameplates_DrawDistance"), self.nMaxRange))
	end
	--Info panel
	if self.bShowHealthMain ~= nil and self.bShowHealthMainDamaged ~= nil then self.wndMain:FindChild("MainShowHealthBarAlways"):SetCheck(self.bShowHealthMain and not self.bShowHealthMainDamaged) end
	if self.bShowHealthMain ~= nil and self.bShowHealthMainDamaged ~= nil then self.wndMain:FindChild("MainShowHealthBarDamaged"):SetCheck(not self.bShowHealthMain and self.bShowHealthMainDamaged) end
	if self.bShowHealthMain ~= nil and self.bShowHealthMainDamaged ~= nil then self.wndMain:FindChild("MainShowHealthBarNever"):SetCheck(not self.bShowHealthMain and not self.bShowHealthMainDamaged) end
	--target components
	if self.bHideInCombat ~= nil then self.wndMain:FindChild("MainHideInCombat"):SetCheck(self.bHideInCombat) end
	if self.bShowMarkerTarget ~= nil then self.wndMain:FindChild("MainHideInCombatOff"):SetCheck(not self.bHideInCombat) end
	--General nameplate occlustion
	self.wndOptionsMain:FindChild("ShowOptionsBacker:MainUseOcclusion"):SetCheck(Apollo.GetConsoleVariable("ui.occludeNameplatePositions"))
end

function Nameplates:OnNormalViewCheck(wndHandler, wndCtrl)
	self.wndMain:FindChild("ContentMain"):Show(true)
	self.wndMain:FindChild("ContentTarget"):Show(false)
end

function Nameplates:OnTargetViewCheck(wndHandler, wndCtrl)
	self.wndMain:FindChild("ContentMain"):Show(false)
	self.wndMain:FindChild("ContentTarget"):Show(true)
end

-- when the OK button is clicked
function Nameplates:OnOK()
	self.wndMain:Show(false) -- hide the window
end


function Nameplates:Firsttime()
	if Allsave["currentspec"] == nil then
		self:SpecChanged()
	end
end

-- when the Cancel button is clicked
function Nameplates:OnCancel()
	self.wndMain:Show(false) -- hide the window
end

function Nameplates:OnDrawDistanceSlider(wndNameplate, wndHandler, nValue, nOldvalue)
	self.wndOptionsMain:FindChild("DrawDistanceLabel"):SetText(String_GetWeaselString(Apollo.GetString("Nameplates_DrawDistance"), nValue))
	self.nMaxRange = nValue-- set new constant, apply math
end

function Nameplates:OnMainShowHealthBarAlways(wndHandler, wndCtrl)
	self:HelperOnMainShowHealthSettingChanged(true, false)
end

function Nameplates:OnMainShowHealthBarDamaged(wndHandler, wndCtrl)
	self:HelperOnMainShowHealthSettingChanged(false, true)
end

function Nameplates:OnMainShowHealthBarNever(wndHandler, wndCtrl)
	self:HelperOnMainShowHealthSettingChanged(false, false)
end

function Nameplates:HelperOnMainShowHealthSettingChanged(bShowHealthMain, bShowHealthMainDamaged)
	self.bShowHealthMain = bShowHealthMain
	self.bShowHealthMainDamaged = bShowHealthMainDamaged
end

function Nameplates:OnMainHideInCombat(wndHandler, wndCtrl)
	self.bHideInCombat = wndCtrl:IsChecked() -- onDraw
end

function Nameplates:OnMainHideInCombatOff(wndHandler, wndCtrl)
	self.bHideInCombat = not wndCtrl:IsChecked() -- onDraw
end

function Nameplates:OnGenericSingleCheck(wndHandler, wndControl, eMouseButton)
	local strSettingName = wndControl:GetData()
	if strSettingName ~= nil then
		self[strSettingName] = wndControl:IsChecked()
		local fnCallback = karSavedProperties[strSettingName].fnCallback
		if fnCallback ~= nil then
			self[fnCallback](self)
		end
	end
end

function Nameplates:OnSettingNameChanged()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		self:DrawNameAndGuild(tNameplate)
		self:DrawLevel(tNameplate)
	end
end

function Nameplates:OnSettingTitleChanged()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		self:DrawNameAndGuild(tNameplate)
	end
end

function Nameplates:OnSettingHealthChanged()
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		self:DrawLevel(tNameplate)
	end
end

function Nameplates:OnOcclusionCheck(wndHandler, wndControl, eMouseButton)
	local bUseOcclusion = wndControl:IsChecked()
	Apollo.SetConsoleVariable("ui.occludeNameplatePositions", bUseOcclusion)
	self.bUseOcclusion = bUseOcclusion
	self.bShowUseOcclusion  = bUseOcclusion
end

function Nameplates:ForceOnOcclusion()
	if self.bShowUseOcclusion ~= nil then
		local bUseOcclusion = self.bShowUseOcclusion
		Apollo.SetConsoleVariable("ui.occludeNameplatePositions", bUseOcclusion)
		self.bUseOcclusion = bUseOcclusion
	end
end

-----------------------------------------------------------------------------------------------
-- Nameplates Instance
-----------------------------------------------------------------------------------------------
local NameplatesInst = Nameplates:new()
NameplatesInst:Init()
